#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use IO::Socket; 
use Fcntl qw(:flock);

# Command line options:
# v=verbose, d=debug,h=help, r=remote host, c=check, i=init, s=sanity check
# n=new tarball, G=generate keys, V=version, N=no mail, C=ICU Config, H=Host config
use vars qw ( $opt_v $opt_d $opt_h $opt_r $opt_c $opt_i 
	      $opt_s $opt_n $opt_G $opt_V $opt_N $opt_C $opt_H );

my $version = 'ICU v0.1 By Andreas stling, andreaso@it.su.se';

# Configuration file to use. 
my $icu_config = './ICU.conf';  

my $verbose = 0;    # 0/1 Always use verbose mode? (Used for testing)
my $debug   = 0;    # 0/1 Always use debug mode?   (Used for testing)

# All sub routines we need
sub parse_command_line;
sub usage;
sub check_root;
sub read_config;
sub check_port;
sub get_host_entry;
sub init_db;
sub create_new_tarball;
sub check_filesystem;
sub getfiles;
sub lockfile;
sub send_mail;
sub get_mail;
sub arrcmp;
sub tmp_dir;
sub clean_exit;
sub notify;
sub beepbeepbeepbeep;
sub icu_die;                 
sub sanity_check;
sub generate_keys;
sub generate_key;

# Hashes we need  
my (%conf, %icuhost);

# Arrays we need 
my (@aide_output,@args);

# Scalars we need
my ( 
     $mail_admin, $check_done, $aide_conf_md5_differs, $scp_output, $i, $init_done, 
     $trying_to_connect, $errormsg, $hostname, $start_date, $subject, $mailbody, $timeout_reason
   ); 

# This must be defined (but the empty value will be overwritten if logger_arguments is set in ICU.conf).
$conf{'logger_arguments'} = "";



#### MAIN ####

$SIG{'ALRM'} = 'beepbeepbeepbeep';

select(STDERR);
$| = 1;
select(STDOUT);
$| = 1;

@args = @ARGV; 	# Save away the original arguments
$start_date = (scalar localtime); 

parse_command_line;
read_config;

if (defined($opt_H)) { 			       # -H <file> => Use this file for host configuration
    $conf{"icuhosts"} = $opt_H;
    $verbose && print "Using $conf{'icuhosts'} as host configuration file.\n";
}

generate_keys if (defined($opt_G));            # Generate keys and exit
sanity_check;				       # Do a sanity check
		
if (get_host_entry($hostname)) {	       # Search for the given hostname in the host configuration file
    $conf{"lockfile_dir"} = $conf{'tmp_dir'};
    $conf{"tmp_dir"} = "$conf{'tmp_dir'}/$icuhost{'name'}";  # Give the host its own temporary directory
    if (defined($opt_n)) {           			     # User specified -n, so create a new tarball
	tmp_dir('new');
	create_new_tarball;
	clean_exit;
    }

    if (defined($opt_i)) {		       # User specified -i, so initialize a new database
	lockfile("create"); 
	tmp_dir("new");
	notify(2,1,1,"Database initialization started");

	if (check_port($icuhost{'name'},$icuhost{'sshport'})) {  
	    init_db;

	    if (defined($mailbody)) {   # Error messages during init should be printed, not sent by mail
		print $mailbody;
		$mailbody = undef;
		$subject  = undef;
	    } else {
		$init_done = 1;		# Mark that the init is done so the welcome mail will be sent later
	    }
	} 
	else {
	    notify(1,1,1,"Could not connect to port $icuhost{'sshport'}"); 
	    get_mail("cant_connect");
	    $subject  .= ": $errormsg";
	    $mailbody .= "Additional information about the connection failure: $errormsg\n"; 			
	}                
	clean_exit;
    } #opt_i

    if (defined($opt_c)) {              # User specified -c so do a filesystem check
	notify(1,1,1,"Check started.");
	lockfile("create");
	tmp_dir("new"); 

	if (check_port($icuhost{'name'},$icuhost{'sshport'})) {
	    check_filesystem;
	} 
	else {
	    notify(1,1,1,"Could not connect to port $icuhost{'sshport'}: $errormsg.");
	    get_mail("cant_connect");
	    $subject  .= ": $errormsg";
	    $mailbody .= "Additional information about the connection failure: $errormsg\n";
	    $mail_admin = 1 if ($conf{'mail_to_icu_admin_port_closed'});
	}

	clean_exit;
    } #opt_c
} 
else {  	# The host was not found
    die("The host $hostname was not found in $conf{'icuhosts'}.\n");
} 

#### END OF MAIN ####



sub parse_command_line	
{
    $debug && print STDERR "* sub parse_command_line started\n";
    my $cmdline_ok = getopts('dvhinsr:cVNC:H:G');

    usage if (defined($opt_h)); 				# -h => show usage and exit
    die("$version\n") if (defined($opt_V)); 			# -V => show version and exit

    usage unless (defined($opt_s) xor defined($opt_n) xor defined($opt_i) 
              xor defined($opt_c) xor defined($opt_G));   	# Only one of them may be specified

    $verbose = 1 if (defined($opt_v) || defined($opt_d) || defined($opt_i)); # Verbose if -v,-d or -i
    $debug   = 1 if (defined($opt_d));    			# -d
    $opt_N   = 1 if (defined($opt_d));                          # Don't sent mail in debug mode 

  # -r must be specified unless running with -s or -G
    if (!defined($opt_r) && !defined($opt_s) && !defined($opt_G)) {
	print STDERR "You must specify the hostname with -r <hostname>\n"; 
	$cmdline_ok = 0;
    } 

    $hostname = $opt_r if (defined($opt_r));			# Grab the hostname

    print "Verbose mode activated.\n" if ($verbose);
    print "Debug mode activated.\n"   if ($debug);

    if (defined($opt_C)) {                   			# Use another ICU configuration file
	$icu_config = $opt_C;
	$verbose && print "Using $icu_config as ICU configuration file.\n";
    }               

    usage unless ($cmdline_ok);  				# Something was wrong

    $debug && print STDERR "* sub parse_command_line ended\n"; 
}                    



sub usage	# Print usage help and exit
{
    select(STDERR);  
    $debug && print STDERR "* sub usage started\n"; 
    print "$version\n";
    print "General usage:\nICU.pl -r <hostname> <action> [options]\n";
    print "Or sanity check:\nICU.pl -s [-r <hostname>] [options]\n";
    print "Or key generation:\nICU.pl -G\n\n";
    print "Actions:\n";
    print "-n                     Create tarball for a new client\n";
    print "-i                     Initialize new database\n";
    print "-c                     Do a filesystem check\n";
    print "Options:\n";
    print "-C <ICU config>        Use <ICU config> as ICU configuration file\n";                    
    print "-H <host config>       Use <host config> as host configuration file\n";
    print "-N                     Print mail to standard output instead of sending it\n"; 
    print "-v                     Verbose mode\n";
    print "-d                     Debug mode - lots of output. AIDE output goes to stdout\n";
    print "Other:\n";
    print "-r <hostname>          Host to work with. Must be one of the hostnames in ICU.hosts\n";
    print "-s                     Just do a sanity check and exit\n";
    print "-G                     Generate keys (should be done only once)\n"; 
    print "-h                     Show this help\n";
    print "-V                     Just show the version and quit\n"; 
    select(STDOUT);
    exit;
}



sub read_config		# Read $icu_config and put parameters and their values in %conf
{
    my (@cfg, @tmparr, $param, $value, $line);
    $debug && print STDERR "* sub read_config started\n"; 
    $debug && print STDERR "Reading ->$icu_config<-\n";

    unless (-e "$icu_config") {
	print STDERR "Fatal error: Could not find $icu_config.\n";
	print STDERR "You must start ICU from the ICU directory or specify which\n";
	print STDERR "configuration file to use with the -C <config> argument.\n";
	exit;
    }

    open(CFG,$icu_config) or die ("CFG: Could not open $icu_config: $!\n");
    @cfg = <CFG>;
    close(CFG);

    for ($i = 0; $i <= $#cfg; $i++) {
	chomp($cfg[$i]);
	$cfg[$i] =~ s/^\s+//;          			 # Remove leading whitespaces
	$cfg[$i] =~ s/\s*\#.*//;       			 # Remove #comments
	$cfg[$i] =~ s/\s+$//; 	       			 # Remove trailing whitespaces
	$cfg[$i] =~ s/\s*=\s*/=/;      			 # Remove whitespaces next to the first =

	($param, $value) = ($cfg[$i] =~ /(.+?)=(.+)/);	 # Try to grab parameter and its value
	if ((defined $param) && (defined $value)) {
	    $param = lc ($param);
	    $conf{"$param"} = $value;			 # Save in %conf
	    $debug && print STDERR "Parameter = ->$param<-, value = ->$conf{$param}<-\n"; 
	}                       
	else {  					 # Not a valid x=y line
	    if ($cfg[$i] =~ /\S+/) {  # don't care about lines with only whitespaces	
		$line = $i + 1;
	        print STDERR "Warning: $icu_config line $line is invalid - ignoring.\n" 
		    if ($verbose || defined($opt_s));
	    }
	}
    } 

  # Replace %icu_basedir% with $conf{'icu_basedir'} in these variables.
    @tmparr = 
      (
         \$conf{'tmp_dir'}, \$conf{'bin_dir'}, \$conf{'config_dir'}, \$conf{'database_dir'},
         \$conf{'key_init_db_pub'}, \$conf{'welcome_mail_body'}, \$conf{'key_init_db_priv'}, 
         \$conf{'key_check_pub'}, \$conf{'key_check_priv'}, \$conf{'key_get_conf_pub'}, 
         \$conf{'key_get_conf_priv'}, \$conf{'key_get_bin_pub'}, \$conf{'key_get_bin_priv'},
         \$conf{'log_dir'}, \$conf{'readme_installscript'}, \$conf{'key_get_db_pub'},
         \$conf{'key_get_db_priv'}, \$conf{'icu_client_installscript'}, \$conf{'icuhosts'} 
      );

    foreach $_(@tmparr) {
	$$_ =~ s/%icu_basedir%/$conf{'icu_basedir'}/ if defined($$_) && defined($conf{'icu_basedir'});
    }

    $debug && print STDERR "* sub read_config ended\n"; 
} 



sub get_host_entry {  # Get host from $conf{"icuhosts"}. 
		      # Syntax: get_host_entry("hostname"). Values will be stored in %icuhost.

    my (@icuhosts, $line);
    my $host = shift;
    $debug && print STDERR "* sub get_host_entry started\n"; 
    $debug && print STDERR "Searching for $host in $conf{'icuhosts'}.\n";

    open(ICUHOSTS,$conf{'icuhosts'}) or die ("ICUHOSTS: Could not open $conf{'icuhosts'}: $!\n");
    @icuhosts = <ICUHOSTS>;
    close(ICUHOSTS);
						
    for ($i = 0; $i <= $#icuhosts; $i++) { 	# Clean up the ICU host configuration file a bit
	$icuhosts[$i] =~ s/^\s+//;          	# Remove leading whitespaces 
	$icuhosts[$i] =~ s/\s*\#.*//;       	# Remove #comments 
	$icuhosts[$i] =~ s/\s*$//; 	    	# Remove trailing whitespaces 

	if ($icuhosts[$i] =~ /^\S+:\S+:\S+:\S+:\d{1,5}$/) {     # Correct host entry 
  	    ($icuhost{"name"},$icuhost{"mailto"},$icuhost{"os"},$icuhost{"arch"}, $icuhost{"sshport"}) 
	      = split(/:/,$icuhosts[$i],5);
		
	    if ($icuhost{'name'} eq $host) {    # It's the host we are looking for
		$debug && print STDERR "* sub get_host_entry returning 1\n";
		return 1;
	    }
	} 
	else {							# Incorrect host entry
            $line = $i + 1;
	    if ($icuhosts[$i] =~ /\S+/) {       # don't care about lines with only whitespaces
	        print STDERR "Warning: $conf{'icuhosts'} line $line is invalid - ignoring.\n" 
		    if ($verbose || defined($opt_s)); 
	    }
	}           
    } 
    undef(%icuhost);
    $debug && print STDERR "* sub get_host_entry returning 0\n";
    return 0;
}



sub init_db    # Initialize a new database
{
    my ($tmpstr, $date, $error);
    $debug && print STDERR "* sub init_db started\n";

  # Run SSH in verbose mode if ICU is running in debug mode
    $conf{"ssh"} =~ s/$conf{'ssh'}/$conf{'ssh'} -v/ if ($debug); 

  # Replace %stuff% with $conf{'stuff'} 
    $conf{"aide_init_db_cmd"} =~ s/%host_basedir%/$conf{'host_basedir'}/g; 	
    $conf{"init_db_cmd"}      =~ s/%key_init_db_priv%/$conf{'key_init_db_priv'}/g;
    $conf{"init_db_cmd"}      =~ s/%port%/$icuhost{'sshport'}/g;	
    $conf{"init_db_cmd"}      =~ s/%hostname%/$icuhost{'name'}/g;
    $conf{"init_db_cmd"}      =~ s/%ssh%/$conf{'ssh'}/;
    $conf{"init_db_cmd"}      =~ s/%aide_init_db_cmd%/\"$conf{'aide_init_db_cmd'}\"/;

  # Now execute the AIDE db init command on the remote host.
  # The right command will be automatically executed on the remote host 
  # because of the SSH key (unless the root acount doesn't have a password).
  # We pass the command to SSH just in case.
    print "Executing init command: '$conf{'init_db_cmd'}'\n";
    print "This may take a while.\n";
    system("$conf{'init_db_cmd'}");

    getfiles("db","conf");
    $date = (scalar localtime);

  # Move aide.db to the host's directory and rename it to aide.db.current
    rename("$conf{'tmp_dir'}/aide.db","$conf{'database_dir'}/$icuhost{'name'}/aide.db.current") 
      || icu_die("$!");

  # Also make a copy of aide.db.current to archive/aide.db-first-$date
    ($error = `$conf{'cp'} \"$conf{'database_dir'}/$icuhost{'name'}/aide.db.current\" \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-first-$date\" 2>&1`) 
      && icu_die("$error"); 
    system("$conf{'compress'} \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-first-$date\" 1>/dev/null") 
      if ($conf{'compress_databases'}); 

  # Move aide.conf to the host's directory and rename it to aide.conf.current
    rename("$conf{'tmp_dir'}/aide.conf","$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current") 
      || icu_die("$!");

  # Also make a copy of aide.conf.current to archive/aide.conf-first-$date
    ($error = `$conf{'cp'} \"$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current\" \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.conf-first-$date\" 2>&1`) 
      && icu_die("$error");

    $debug && print STDERR "* sub init_db ended\n";
}



sub create_new_tarball		# Create client tarball
{
    my ( @instscript, $tmpstr, $key_check_pub_path, $key_init_db_pub_path,
         $key_get_bin_pub_path, $key_get_conf_pub_path, $key_get_db_pub_path );

    $debug && print STDERR "* sub create_new_tarball started\n";

  # Does the binary exist?
    $tmpstr = "$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}";
    die ("Fatal error, Could not find the binary $tmpstr (You have not compiled AIDE for $icuhost{'os'}/$icuhost{'arch'}?)\n") 
      unless (-e "$tmpstr");

  # Does the configuration file exist?
    $tmpstr = "$conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'}";
    die ("Fatal error, Could not find the configuration file $tmpstr. (You have not made any default configuration file for $icuhost{'os'}/$icuhost{'arch'}?)\n")
      unless (-e "$tmpstr");

    notify(2,1,1,"Creating new tarball for $icuhost{'name'}.");

  # Create temporary directory where we keep the files that should be in the tarball
    $tmpstr = "$conf{'tmp_dir'}/$icuhost{'name'}.icu-install";
    print "Creating directory $tmpstr.\n";
    mkdir("$tmpstr",0700) or die ("Fatal error: Could not create $tmpstr: $!\n\nExiting..."); 

  # Create directories where the client's databases/configurations etc will be kept
    $tmpstr = "$conf{'database_dir'}/$icuhost{'name'}";
    mkdir("$tmpstr",0700) or die("Fatal error: Could not create $tmpstr: $!\n\nExiting...");
    $tmpstr .= "/archive";
    mkdir("$tmpstr",0700) or die("Fatal error: Could not create $tmpstr: $!\n\nExiting..."); 

  # Copy binary and configuration to the temporary directory.
    system("$conf{'cp'} \"$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/aide.bin\"");
    system("$conf{'cp'} \"$conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'}\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/aide.conf\"");

  # Copy the init_db key
    $key_init_db_pub_path = $conf{'key_init_db_pub'};    # Path and filename here
    $conf{'key_init_db_pub'} =~ s/.+\///;                # Only the filename here
    system("$conf{'cp'} \"$key_init_db_pub_path\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/$conf{'key_init_db_pub'}\"");

  # Copy the check key
    $key_check_pub_path = $conf{'key_check_pub'};
    $conf{'key_check_pub'} =~ s/.+\///;
    system("$conf{'cp'} \"$key_check_pub_path\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/$conf{'key_check_pub'}\"");

  # Copy the get_bin key
    $key_get_bin_pub_path = $conf{'key_get_bin_pub'};
    $conf{'key_get_bin_pub'} =~ s/.+\///;
    system("$conf{'cp'} \"$key_get_bin_pub_path\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/$conf{'key_get_bin_pub'}\"");

  # Copy the get_conf key
    $key_get_conf_pub_path = $conf{'key_get_conf_pub'};
    $conf{'key_get_conf_pub'} =~ s/.+\///;
    system("$conf{'cp'} \"$key_get_conf_pub_path\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/$conf{'key_get_conf_pub'}\""); 

  # Copy the get_db key
    $key_get_db_pub_path = $conf{'key_get_db_pub'};
    $conf{'key_get_db_pub'} =~ s/.+\///;
    system("$conf{'cp'} \"$key_get_db_pub_path\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/$conf{'key_get_db_pub'}\"");

  # Copy the install script README
    system("$conf{'cp'} \"$conf{'readme_installscript'}\" \"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/README\"");

  # Open the install script and replace %stuff% with $conf{'stuff'}
    open(INSTSCRIPT,$conf{'icu_client_installscript'}) or die "INSTSCRIPT: Could not open $conf{'icu_client_installscript'}: $!\n";
    @instscript = <INSTSCRIPT>;
    close(INSTSCRIPT);

    $conf{'aide_init_db_cmd'} =~ s/%host_basedir%/$conf{'host_basedir'}/g;
    for ($i = 0; $i <= $#instscript; $i++) {
	$instscript[$i] =~ s/%host_basedir%/$conf{'host_basedir'}/g;		
	$instscript[$i] =~ s/%key_check_pub%/$conf{'key_check_pub'}/g; 
	$instscript[$i] =~ s/%key_init_db_pub%/$conf{'key_init_db_pub'}/g;
	$instscript[$i] =~ s/%key_get_bin_pub%/$conf{'key_get_bin_pub'}/g;  
	$instscript[$i] =~ s/%key_get_conf_pub%/$conf{'key_get_conf_pub'}/g; 
	$instscript[$i] =~ s/%key_get_db_pub%/$conf{'key_get_db_pub'}/g; 
	$instscript[$i] =~ s/%icu_server_name%/$conf{'icu_server_name'}/g;
	$instscript[$i] =~ s/%admin_email%/$conf{'admin_email'}/g;
	$instscript[$i] =~ s/%hostname%/$icuhost{'name'}/g;
	$instscript[$i] =~ s/%arch%/$icuhost{'arch'}/g;  
	$instscript[$i] =~ s/%os%/$icuhost{'os'}/g;
    }

  # Write the install script to disk.
    $tmpstr = "$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/icu-install.sh";
    open(INSTSCRIPT,">$tmpstr") or die "INSTSCRIPT: Could not open $tmpstr for writing: $!\n";
    print INSTSCRIPT @instscript;
    close(INSTSCRIPT);
    chmod(0700,"$conf{'tmp_dir'}/$icuhost{'name'}.icu-install/icu-install.sh");

  # Now create the tarball.
    chdir("$conf{'tmp_dir'}") or die("Fatal error, could not change to working directory $conf{'tmp_dir'}: $!\n");
    $tmpstr = "$conf{'database_dir'}/$icuhost{'name'}/$icuhost{'name'}.icu-install.tar";
    system("$conf{'tar'} cf $tmpstr $icuhost{'name'}.icu-install");

  # Make sure the tarball was actually created
    if (-e "$tmpstr") {	 
	notify(2,1,1,"Saved tarball as $tmpstr.");
	print "This file must now be untared and installed on the host $icuhost{'name'}.\n";
    }
    else {
	print STDERR "Something went wrong - could not create tarball.\n";
	print STDERR "Inspect previous error messages and try again.\n";
    }

    $debug && print STDERR "* sub create_new_tarball ended\n"; 
}



sub check_filesystem	# Do a filesystem check
{
    my ( @arr1, @arr2, @files, @missing_files, $tmpstr, $date, $local_md5sum, $error,
         $remote_md5sum,$added, $removed, $mailbody_temp, $subject_temp );

    $debug && print STDERR "* sub check_filesystem started\n"; 

  # Get binary, configuration and database from the remote host so we can verify the MD5 fingerprints.
    getfiles("bin","conf","db");    # Get aide, aide.conf and aide.db.

  # Verify MD5 fingerprint of aide.db.
    $verbose && print "Verifying MD5 fingerprint of the AIDE database...";
    $_ = `$conf{'md5sum'} $conf{'database_dir'}/$icuhost{'name'}/aide.db.current`;
    $_ =~ /([\da-f]{32})/;	# Extract MD5 fingerprint from the output
    $local_md5sum = $1;
    $_ = `$conf{'md5sum'} $conf{'tmp_dir'}/aide.db`;
    $_ =~ /([\da-f]{32})/;
    $remote_md5sum = $1;

    if ($local_md5sum eq $remote_md5sum) {              # MD5 ok
	$verbose && print "match.\n";
    } 
    else {						# MD5 NOT ok
	$date = (scalar localtime);
	$verbose && print "MISMATCH.\n"; 
	notify(1,1,1,"The MD5 fingerprint of aide.db did NOT match!");

  # Incorrect MD5 fingerprint - Check what has been added/removed.
	$tmpstr = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
	open(OLDDB,"$tmpstr") or icu_die ("OLDDB: Could not open $tmpstr: $!");
	@arr1 = <OLDDB>;
	close(OLDDB);

	$tmpstr = "$conf{'tmp_dir'}/aide.db";
	open(NEWDB,"$tmpstr") or icu_die ("NEWDB: Could not open $tmpstr: $!");
	@arr2 = <NEWDB>;
	close(NEWDB);

	get_mail("aide_db_md5_mismatch");
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_db_md5'});  

        $mailbody .= "Additional information:\n";
        $mailbody .= "Fingerprint of locally stored database: $local_md5sum\n";
        $mailbody .= "Fingerprint of remote database:         $remote_md5sum\n";
 
  # Print size of locally stored database:
        ($tmpstr) = (stat("$conf{'database_dir'}/$icuhost{'name'}/aide.db.current"))[7];
        $mailbody .= "Size of locally stored database: $tmpstr\n";
 
  # Print size of remote database:
        ($tmpstr) = (stat("$conf{'tmp_dir'}/aide.db"))[7];
        $mailbody .= "Size of remote database:        $tmpstr\n";

  # Update database/hostname/aide.db.current
        notify(1,1,1,"Updating $conf{'database_dir'}/$icuhost{'name'}/aide.db.current");
        rename("$conf{'tmp_dir'}/aide.db","$conf{'database_dir'}/$icuhost{'name'}/aide.db.current") || icu_die("$!");
        notify(1,1,1,"Saving copy as $conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date.");
        ($error = `$conf{'cp'} \"$conf{'database_dir'}/$icuhost{'name'}/aide.db.current\" \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date\" 2>&1`)
          && icu_die("$error");
        system("$conf{'compress'} \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date\" 1>/dev/null")
          if ($conf{'compress_databases'});

	$verbose && print "Finding differences between databases...";
	($added,$removed) = arrcmp(\@arr1,\@arr2);  # Find differences between @arr1 and @arr2
	$verbose && print "done.\n";

	if (defined($added)) {			# Something had been added to aide.db
	    $mailbody .= "\nThe following line(s) had been added to $conf{'host_basedir'}/aide.db ";
	    $mailbody .= "on $icuhost{'name'}:\n$added\n";
	}

	if (defined($removed)) { 		# Something had been removed from aide.db
	    $mailbody .= "\nThe following line(s) had been removed from $conf{'host_basedir'}/aide.db "; 
	    $mailbody .= "on $icuhost{'name'}:\n$removed\n";  
	}
    }

  # Verify the MD5 fingerprint of aide.conf. 
    $verbose && print "Verifying MD5 fingerprint of the AIDE configuration..."; 
    $_ = `$conf{'md5sum'} $conf{'database_dir'}/$icuhost{'name'}/aide.conf.current`;
    $_ =~ /([\da-f]{32})/;
    $local_md5sum = $1;
    $_ = `$conf{'md5sum'} $conf{'tmp_dir'}/aide.conf`;
    $_ =~ /([\da-f]{32})/;
    $remote_md5sum = $1;       
	
    if ($local_md5sum eq $remote_md5sum) {   
	$verbose && print "match.\n";           # MD5 ok
    } 
    else {					# MD5 NOT ok
	$verbose && print "MISMATCH.\n";       
	$date = (scalar localtime); 
	$aide_conf_md5_differs = 1;
	notify(1,1,1,"The MD5 fingerprint of aide.conf did NOT match!");

  # Incorrect MD5 fingerprint - Check what has been added/removed.    
	$tmpstr = "$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current";
        open(OLDCONF,"$tmpstr") or icu_die ("OLDCONF: Could not open $tmpstr: $!");
       	@arr1 = <OLDCONF>;
       	close(OLDCONF);

	$tmpstr = "$conf{'tmp_dir'}/aide.conf";
        open(NEWCONF,"$tmpstr") or icu_die ("NEWCONF: Could not open $tmpstr: $!");
       	@arr2 = <NEWCONF>;
       	close(NEWCONF);

	get_mail("aide_conf_md5_mismatch");
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_conf_md5'});

        $mailbody .= "Additional information:\n";
        $mailbody .= "Fingerprint of locally stored configuration:  $local_md5sum\n";
        $mailbody .= "Fingerprint of remote configuration:          $remote_md5sum\n";
 
  # Print size of locally stored config:
        ($tmpstr) = (stat("$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current"))[7];
        $mailbody .= "Size of locally stored configuration: $tmpstr\n";
 
  # Print size of remote config:
        ($tmpstr) = (stat("$conf{'tmp_dir'}/aide.conf"))[7];
        $mailbody .= "Size of remote configuration:         $tmpstr\n";

  # Update database/hostname/aide.conf.current
        notify(1,1,1,"Updating $conf{'database_dir'}/$icuhost{'name'}/aide.conf.current");
        rename("$conf{'tmp_dir'}/aide.conf","$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current")
          || icu_die("$!");
        notify(1,1,1,"Saving copy as $conf{'database_dir'}/$icuhost{'name'}/archive/aide.conf-$date.");
        ($error = `$conf{'cp'} \"$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current\" \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.conf-$date\" 2>&1`)
          && icu_die("$error");

	$verbose && print "Finding differences between configurations..."; 
	($added,$removed) = arrcmp(\@arr1,\@arr2);
	$verbose && print "done.\n";

	if (defined($added)) {                  # Something had been added to aide.conf
	    $mailbody .= "\nThe following line(s) had been added to $conf{'host_basedir'}/aide.conf ";
	    $mailbody .= "on $icuhost{'name'}:\n$added\n";
	}

	if (defined($removed)) {                # Something had been removed from aide.conf
	    $mailbody .= "\nThe following line(s) had been removed from $conf{'host_basedir'}/aide.conf ";
	    $mailbody .= "on $icuhost{'name'}:\n$removed\n";
	}
    }                

  # Verify the MD5 fingerprint of the AIDE binary.
    $verbose && print "Verifying MD5 fingerprint of the AIDE binary...";
    $_ = `$conf{'md5sum'} $conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}`;
    $_ =~ /([\da-f]{32})/;
    $local_md5sum = $1;
    $_ = `$conf{'md5sum'} $conf{'tmp_dir'}/aide.bin`;
    $_ =~ /([\da-f]{32})/;
    $remote_md5sum = $1;         

    if ($local_md5sum eq $remote_md5sum) {         # MD5 ok
	$verbose && print "match.\n";
    } 
    else {                                         # MD5 NOT ok
	$verbose && print "MISMATCH.\n";
	notify(1,1,1,"The MD5 fingerprint of the AIDE binary did NOT match!");
        $date = (scalar localtime);

	get_mail("aide_binary_md5_mismatch");
	$mail_admin = 1; # always send mail to admin when the binary has changed

	$mailbody .= "Additional information:\n";
	$mailbody .= "Fingerprint of locally stored binary:  $local_md5sum\n";
	$mailbody .= "Fingerprint of remote binary:          $remote_md5sum\n";

  # Print size of locally stored binary:
        ($tmpstr) = (stat("$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}"))[7];
        $mailbody .= "Size of locally stored binary: $tmpstr\n";

  # Print size of remote binary:
	($tmpstr) = (stat("$conf{'tmp_dir'}/aide.bin"))[7]; 
	$mailbody .= "Size of remote binary:         $tmpstr\n";

        notify(1,1,1,"Saving remote mismatching AIDE binary to $conf{'database_dir'}/$icuhost{'name'}/archive/aide.bin-md5-mismatch-$date.");
          ($error = `$conf{'cp'} \"$conf{'tmp_dir'}/aide.bin\" \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.bin-md5-mismatch-$date\" 2>&1`)
          && icu_die("$error");

  # Should be continue even if the binary has changed?!
	if ($conf{'continue_if_binary_changed'}) {
	    $subject .= " - Check done anyway."; 
	    notify(1,1,1,"WARNING, continuing anyway.");
        } 
	else {
	    notify(1,1,1,"Check aborted.");
	    clean_exit;
	}
    }
                                                                                               
    $subject_temp  = $subject;   # Save away old mail
    $mailbody_temp = $mailbody;

    get_mail("check_timeout");   # If the check times out, this mail will be sent
    $mailbody .= "Additional information: Check timed out after $conf{'check_timeout'} seconds.\n";
    $debug && print STDERR "Setting timeout to $conf{'check_timeout'} seconds.\n";
    $timeout_reason = "Check timed out after $conf{'check_timeout'} seconds.\n";
    alarm($conf{'check_timeout'});
	
  # Run SSH in verbose mode if ICU is running in debug mode
    $conf{"ssh"} =~ s/$conf{'ssh'}/$conf{'ssh'} -v/ if ($debug);  
  # Run SSH in quiet mode if ICU isn't running in debug or verbose mode
    $conf{"ssh"} =~ s/$conf{ssh}/$conf{ssh} -q/ unless ($debug || $verbose);

  # Replace %stuff% with $conf{'stuff'} in the check commands
    $conf{"aide_check_cmd"}  =~ s/%host_basedir%/$conf{'host_basedir'}/g;
    $conf{"check_cmd"}       =~ s/%key_check_priv%/$conf{'key_check_priv'}/g;
    $conf{"check_cmd"}       =~ s/%port%/$icuhost{'sshport'}/g;
    $conf{"aide_check_cmd"}  =~ s/%host_basedir%/$conf{'host_basedir'}/g;
    $conf{"check_cmd"}       =~ s/%hostname%/$icuhost{'name'}/g; 
    $conf{"check_cmd"}       =~ s/%ssh%/$conf{'ssh'}/;               
    $conf{"check_cmd"}       =~ s/%aide_check_cmd%/\"$conf{'aide_check_cmd'}\"/;

  # Now execute the real check.
    $debug && print STDERR "Executing $conf{'check_cmd'}\n"; 
    $verbose && print "Executing AIDE check command on the remote host. This may take a while.\n";
    @aide_output = "";

    if ($debug) {			  # debug mode => give output in realtime 
	system("$conf{'check_cmd'}");         
    }
    else {			          # not debug mode => save (possible) output for later use
	open(CHECK,"$conf{'check_cmd'}|");
        while ($_ = <CHECK>) {
            push(@aide_output ,$_) 
	    unless ($_ =~ /no such file or directory/i);  # Ignore warnings about missing files/directories
        }
        close(CHECK);
    }                    

  # The check didn't time out so stop the alarm clock and restore the original mail
    alarm(0);
    $mailbody       = $mailbody_temp;
    $subject  	    = $subject_temp;
    $timeout_reason = undef;
    $check_done = 1;

    if ($#aide_output == 0) {              # No change was found
	if ($conf{'allow_empty_reports'}) {
	    get_mail("empty_report");
	} 
	else {
	    notify(1,1,1,"No change in the filesystem was found.");
	}
    } 
    else {			           # A change was found
	get_mail("filesystem_changed");  
	foreach $_ (@aide_output) {     # Show summary in the subject.
	    $subject .= "(Added=$1,"    if ($_ =~ /added files=(\d+)/i);
	    $subject .= "Removed=$1,"   if ($_ =~ /removed files=(\d+)/i); 
	    $subject .= "Changed=$1) "  if ($_ =~ /changed files=(\d+)/i); 
        }
	$mailbody .= "Below is the output from AIDE. Read it carefully.\n\n@aide_output";
    }                                              

  # Always get the new database (the timestamp in it always changes after a check)   
    getfiles("db");

  # Update aide.db on the ICU server. 
    rename("$conf{'tmp_dir'}/aide.db","$conf{'database_dir'}/$icuhost{'name'}/aide.db.current") 
      || icu_die("$!");

  # Also save a copy of aide.db in the archive directory if a change in the filesystem was found
    unless ($#aide_output == 0) { 
	$date = (scalar localtime);
	notify(1,1,1,"A change in the filesystem was found, updating $conf{'database_dir'}/$icuhost{'name'}/aide.db.current."); 
	notify(1,1,1,"Saving copy as $conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date.");
	($error = `$conf{'cp'} \"$conf{'database_dir'}/$icuhost{'name'}/aide.db.current\" \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date\" 2>&1`) 
	  && icu_die("$error"); 
	system("$conf{'compress'} \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date\" 1>/dev/null")
	  if ($conf{'compress_databases'});
    }

    $debug && print STDERR "* sub check_filesystem ended\n";
}



sub getfiles   # Get file(s) from the remote host.  Syntax: getfiles("db"||"conf"||"bin")
{
    my (@tmparr, @files, @missing_files, $mailbody_temp, $subject_temp);
    my $get_db   = 0;
    my $get_bin  = 0;
    my $get_conf = 0;

    $debug && print STDERR "* sub getfiles started\n";

    $get_bin = 1 				# Should we get the binary?
        if (defined($_[0]) && $_[0] eq "bin" 
         || defined($_[1]) && $_[1] eq "bin" 
         || defined($_[2]) && $_[2] eq "bin");
  
    $get_conf = 1 				# Should we get the config?
	if (defined($_[0]) && $_[0] eq "conf" 
	 || defined($_[1]) && $_[1] eq "conf" 
	 || defined($_[2]) && $_[2] eq "conf");

    $get_db = 1					# Should we get the database?
	if (defined($_[0]) && $_[0] eq "db" 
	 || defined($_[1]) && $_[1] eq "db" 
	 || defined($_[2]) && $_[2] eq "db"); 

    if ($debug) {				# Run scp in verbose mode if ICU is running in debug mode 
	$conf{"scp"} =~ s/$conf{'scp'}/$conf{'scp'} -v/; # Run scp in verbose mode if ICU is running in debug mode
    } 
    else {					# Run scp in quiet mode unless ICU is running in debug or init mode 
	$conf{"scp"} =~ s/$conf{'scp'}/$conf{'scp'} -q/ unless (defined($opt_i));
    }

  # Replace %stuff% with $conf{'stuff'} in the get* commands
    @tmparr = (\$conf{'get_bin_cmd'}, \$conf{'get_conf_cmd'}, \$conf{'get_db_cmd'});
    foreach $_(@tmparr) {
	$$_ =~ s/%scp%/$conf{'scp'}/; 
	$$_ =~ s/%port%/$icuhost{'sshport'}/;
	$$_ =~ s/%host_basedir%/$conf{'host_basedir'}/g; 
	$$_ =~ s/%tmp_dir%/$conf{'tmp_dir'}/g;
	$$_ =~ s/%key_get_bin_priv%/$conf{'key_get_bin_priv'}/g; 
	$$_ =~ s/%key_get_conf_priv%/$conf{'key_get_conf_priv'}/g;
	$$_ =~ s/%key_get_db_priv%/$conf{'key_get_db_priv'}/g;
	$$_ =~ s/%hostname%/$icuhost{'name'}/g; 
    }

    tmp_dir("new");
    $debug && print STDERR "Setting timeout to $conf{'getfiles_timeout'} seconds.\n";

    $subject_temp = $subject;
    $mailbody_temp = $mailbody;
    get_mail("getfiles_timeout");
    $mailbody .= "Additional information: File transfer timed out after $conf{'getfiles_timeout'} seconds.\n";
    $subject  .= "- File transfer timed out after $conf{'getfiles_timeout'} seconds.";
    @tmparr = "";   
    $scp_output = undef;
    $timeout_reason = "File transfer timed out after $conf{'getfiles_timeout'} seconds.\n";
    alarm($conf{'getfiles_timeout'});  

    $verbose && print "Getting files from $icuhost{'name'}: ";

  # Get aide.bin
    if ($get_bin) {
	if ($verbose) {
	    print "aide.bin ";
  	    $debug && print STDERR "\nExecuting $conf{'get_bin_cmd'}\n";
	    system("$conf{'get_bin_cmd'}");     # verbose mode => give (possible) output in realtime
	} 
	else {
	open(GET,"$conf{'get_bin_cmd'}|");      # not verbose mode => save (possible) output for later use
	    while ($_ = <GET>) {
		$scp_output .= $_ unless ($_ =~ /executing scp1 compatibility/i);  # comp. warnings are no fun 
	    }
	    close(GET);
	}
    }

  # Get aide.conf
    if ($get_conf) {
	if ($verbose) {
	    print "aide.conf ";
	    $debug && print STDERR "\nExecuting $conf{'get_conf_cmd'}\n";
            system("$conf{'get_conf_cmd'}");
        } 
	else {
            open(GET,"$conf{'get_conf_cmd'}|");
            while ($_ = <GET>) {
		$scp_output .= $_ unless ($_ =~ /executing scp1 compatibility/i);
            }
            close(GET);
        }
    }                     

  # get aide.db
    if ($get_db) {
        if ($verbose) {
	    print "aide.db ";
	    $debug && print STDERR "\nExecuting $conf{'get_db_cmd'}\n";
            system("$conf{'get_db_cmd'}");
        } 
        else {
  	    open(GET,"$conf{'get_db_cmd'}|");
	    while ($_ = <GET>) {
		$scp_output .= $_ unless ($_ =~ /executing scp1 compatibility/i);
	    }
            close(GET);
        }
    }                     
    print "\n" unless (defined($opt_i) || !$verbose);
    alarm(0);
    $mailbody       = $mailbody_temp;
    $subject        = $subject_temp;
    $timeout_reason = undef;

# if scp gave any output - send it in the mail and also write it to the host's log.
    if (defined($scp_output)){  
	notify(1,1,1,"Warning: scp gave output when copying files.");
        notify(0,0,1,"Here is the output scp gave:\n$scp_output");
        notify(0,0,1,"End of scp output.\n");
	get_mail("scp_output") if (defined($scp_output));
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_scp_output'});   
    }

  # Check if the required files was actually received.
    push(@files,"$conf{'tmp_dir'}/aide.bin")  if ($get_bin);	
    push(@files,"$conf{'tmp_dir'}/aide.conf") if ($get_conf);	
    push(@files,"$conf{'tmp_dir'}/aide.db")   if ($get_db);
    @missing_files = "";

    for ($i = 0; $i <= $#files; $i++) {
	if (-e "$files[$i]") {         
	    $debug && print STDERR "$files[$i] was successfully received.\n";
	} 
	else { 
	    notify(1,1,1,"$files[$i] was NOT received.");
	    $files[$i] =~ s/.+\///;  
	    $files[$i] = "$files[$i] ($conf{'host_basedir'}/$files[$i] on $icuhost{'name'})";
	    push(@missing_files,"$files[$i]\n");
	}
    }
	
    if ($#missing_files > 0) {
	notify(1,1,1,"Critical file(s) missing, aborting.");
	get_mail("files_missing");
	$mailbody .= "Note: This was after the actual check.\n" if ($check_done);
	$mailbody .= "Note: This was before the actual check.\n" unless ($check_done);
	$mailbody .= "Either the file transfer failed or the files were missing on the remote host.\n";
	$mailbody .= "Here is the list of missing files:\n@missing_files\n";
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_missing_files'});
	clean_exit;
    } 
    else {
 	$verbose && print "All files successfully received.\n";
    }

    $debug && print STDERR "* sub getfiles ended\n";
}



sub lockfile    # Create or remove lock file.  Syntax: lockfile("create") or lockfile("remove");
{
    my ($lockfile, $mailto);	
    $debug && print STDERR "* sub lockfile started\n";

    my $action = shift;	  # action is lock or remove
    $lockfile  = "$conf{'lockfile_dir'}/$icuhost{'name'}.lock";
    $mailto    = $conf{'admin_email'};	

#  Create new lock file
    if ($action eq "create") {
	$debug && print STDERR "Checking lock file.\n";

	if (-e "$lockfile") {      # Does a lock file already exist?
	    notify(1,1,1,"Lock file found for $icuhost{'name'}: $lockfile\nCheck/init already in progress?\nAborting.");
	    get_mail("lockfile_exists") unless (defined($opt_i)); 
	    if (defined($mailbody)) {
	        if (defined($opt_N)) {
		    print $mailbody;
		} 
		else {
		    $mailto .= ",$icuhost{'mailto'}" if ($conf{'lockfile_also_notify_host_admin'});
		    send_mail($mailto,0,$subject,$mailbody);
	        }                                   
	    }
	    exit;
        } 
	else {	# Create new lock file
	    $debug && print STDERR "Not found. Creating new.\n";
	    open(LOCKFILE,">$lockfile") or icu_die("Could not create lockfile $lockfile: $!\n");
	    print LOCKFILE "This is a lock file for the host $icuhost{'name'}.\n";
	    print LOCKFILE "Process started $start_date, pid $$.\n";
	    print LOCKFILE "Command line: ICU @args\n";
	    close(LOCKFILE);
	}
    } 
  # Remove lock file
    elsif ($action eq "remove") {
        $debug && print STDERR "Removing lock file ($lockfile)\n";
        if (-e "$lockfile") {  
   	    unlink("$lockfile") or icu_die("Could not remove $lockfile: $!\n");
	}		
    }

    $debug && print STDERR "* sub lockfile ended\n";
}



sub check_root      # Check if ICU is started as root. 
{
    my $timeout;
    $debug && print STDERR "* sub check_root started\n";
    unless ($>) { 
	$timeout = 20;
        print STDERR "You are running ICU.pl as user root and that's probably not what you want.\n";
        print STDERR "Press CTRL-C within $timeout seconds to quit, or just wait to continue.\n";
        sleep $timeout;
	print "Continuing...\n";
    }
    $debug && print STDERR "* sub check_root ended\n";
}



sub check_port     # Do a TCP connect to check if given port is open on a host. Syntax: check_port(hostname,port)
{
    my ($hostname, $port, $iaddr, $paddr, $proto);
    $debug && print STDERR "* sub check_port started\n";

    $hostname = shift;
    $port     = shift;
    $iaddr    = inet_aton($hostname);
    $paddr    = sockaddr_in($port,$iaddr);
    $proto    = getprotobyname('TCP');   

    socket(SOCK,PF_INET,SOCK_STREAM,$proto);
    $verbose && print "Checking if port $port is open.\n";

# Get the "can't connect" mail just in case it times out.
# $mailbody and $subject are cleared if the connection is successful.  
    get_mail("cant_connect");
    $mailbody .= "Additional information: Connection timed out after $conf{'check_port_timeout'} seconds.\n";
    $subject  .= ": Connection timed out after $conf{'check_port_timeout'} seconds.";
    $trying_to_connect = 1;  # So the beepbeepbeepbeep sub knows what we were trying to do
    $timeout_reason = "Could not connect to port $port: Connection timed out after $conf{'check_port_timeout'} seconds.\n";
    alarm($conf{'check_port_timeout'});

    if (connect(SOCK,$paddr)) {             # The port is open
	alarm(0);
	$trying_to_connect = 0;
	$debug && print STDERR "Port is open.\n";
	$timeout_reason = undef;
	$mailbody       = undef;
	$subject        = undef;
	close(SOCK);
	$debug && print STDERR "* sub check_port returning 1\n";
	return 1;
    } 
    else {				    # The port is closed
	alarm(0);
	$trying_to_connect = 0;
	$mailbody       = undef;
	$subject        = undef;
	$timeout_reason = undef;
	$errormsg = $!;
	close(SOCK);
	$debug && print STDERR "* sub check_port returning 0\n";
	return 0;
    }

    $debug && print STDERR "* sub check_port ended\n";
}



sub send_mail   #  Syntax: send_mail(to,0/1,subject,body)   
		#  0 = Don't send copy to ICU admin, 1 = Also send copy to ICU admin. 
{
    my (@body, $mailto, $to_icu_admin, $subject, $tmpstr, $date);	

    $debug && print STDERR "* sub send_mail started\n";
    $mailto       = shift;
    $to_icu_admin = shift; 
    $subject      = shift;
    @body         = @_;
    $mailto      .= ",$conf{'admin_email'}" if ($to_icu_admin);	# Also send to ICU admin?
	
  # Send empty report?
    if ($check_done && $#aide_output == 0 && !$conf{'allow_empty_reports'}) {
 	push(@body,"\nThe AIDE filesystem check was executed as normal, and found no change in the filesystem.\n");
    }

    unless ($check_done || defined($opt_i)) {	# Give an extra note if no check has been done
	$_ = "[THE CHECK WAS ABORTED!!]";
	$subject .= " $_";
	push(@body,"\n$_\n");
    }

    $subject = "[ICU - $hostname]  $subject";
    notify(1,1,1,"Sending mail to $mailto with subject: $subject");
    $date = (scalar localtime);
    $tmpstr = "$conf{'log_dir'}/$icuhost{'name'}.log";

    unless (defined($opt_i)) {		# Save the mail in the hosts' log (but not the welcome mail).
	notify(0,0,1,"Message body:\n@body");
	notify(0,0,1,"End of message body.");
    }

    open(SENDMAIL, "|$conf{'sendmail'} $mailto") or die ("SENDMAIL: Could not open sendmail process: $!\n");
    print SENDMAIL "From: $conf{'mail_from'} <$conf{'admin_email'}>\n";
    print SENDMAIL "Subject: $subject\n";
    print SENDMAIL "@body\n\n";
    close(SENDMAIL);

    $debug && print STDERR "* sub send_mail ended\n";
}


sub get_mail      # Syntax: get_mail("mail")  Result will be appended to $mailbody and $subject
{
    $debug && print STDERR "* sub get_mail started\n";
    
    my $mailtype = shift;
    $debug && print STDERR "Getting mail ->$mailtype<-\n";

    if ($mailtype eq "cant_connect") {
	$subject  .= "Warning: Could not connect to port $icuhost{'sshport'}";      
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date was aborted\n";
        $mailbody .= "because the ICU server ($conf{'icu_server_name'}) was unable to connect to port\n";
	$mailbody .= "$icuhost{'sshport'} on $icuhost{'name'}. This could mean that $icuhost{'name'}\n";
        $mailbody .= "isn't running sshd on port $icuhost{'sshport'} or that the host is down.\n";
	$mailbody .= "You should investigate this as soon as possible.\n";
    }

    if ($mailtype eq "filesystem_changed") { 
	$subject  .= "Warning: Filesystem has changed "; 
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The filesystem on $icuhost{'name'} has changed.\n";  
	$mailbody .= "This could mean that authorized changes were made, but it could\n";
	$mailbody .= "also mean that the host has been compromised. The database has been\n";  
	$mailbody .= "updated with these changes and will now be regarded as safe.\n"; 
    }

    if ($mailtype eq "aide_binary_md5_mismatch") {
	$subject  .= "Warning: The AIDE BINARY has changed!! ";
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE binary ($conf{'host_basedir'}/aide.bin) on $icuhost{'name'}\n";
        $mailbody .= "differs from the AIDE binary stored locally on the ICU server.\n";
	$mailbody .= "THIS SHOULD NEVER HAPPEN and could mean that the host has been\n"; 
	$mailbody .= "compromised. Investigate this as soon as possible.\n";
    }
	
    if ($mailtype eq "aide_db_md5_mismatch") {
        $subject  .= "Warning: The AIDE database file has changed. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE database file on $icuhost{'name'} differs from\n"; 
	$mailbody .= "the database file stored locally on the ICU server.\n";
        $mailbody .= "This could mean that authorized changes were made,\n";
        $mailbody .= "but it could also mean that the host has been compromised.\n";
	$mailbody .= "The AIDE database on the ICU server has been updated with these changes\n";
	$mailbody .= "and will now be regarded as safe. If only the timestamp has changed\n";
	$mailbody .= "it might mean that a check was run locally or that an earlier ICU check";
        $mailbody .= "was aborted for some reason.\n";
    }

    if ($mailtype eq "aide_conf_md5_mismatch") {    
	$subject  .= "Warning: The AIDE configuration file has changed. ";  
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE configuration file on $icuhost{'name'} differs from\n";
	$mailbody .= "the configuration file stored locally on the ICU server.\n";
	$mailbody .= "This could mean that authorized changes were made, \n";
	$mailbody .= "but it could also mean that the host has been compromised.\n";
	$mailbody .= "The AIDE configuration file on the ICU server has been updated with these\n"; 
	$mailbody .= "changes and will now be regarded as safe.\n"; 
    }

    if ($mailtype eq "files_missing") {
	$subject  .= "Warning: Critical file(s) missing. ";
        $mailbody .= "\n*** Warning ***\n";
        $mailbody .= "At least one critical file was missing on $icuhost{'name'}.\n";
    }                                                                    

    if ($mailtype eq "getfiles_timeout") {
	$subject  .= "Warning: Could not get files ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date was aborted\n";
        $mailbody .= "because the file transfer timed out.\n";
        $mailbody .= "You should investigate this as soon as possible.\n";
    }                            

    if ($mailtype eq "check_timeout") {
	$subject  .= "Warning: The AIDE filesystem check timed out. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date timed out after\n";
	$mailbody .= "You shold investigate this as soon as possible.\n";
    }                         

    if ($mailtype eq "lockfile_exists") {
        $subject  .= "Warning: Lock file found while trying to initialize the filesystem check. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date was aborted because\n";
        $mailbody .= "a lock file was found. A check may already be in progress or a fatal error\n";
	$mailbody .= "occured during an earlier check. The ICU administrator (who has now been\n";
	$mailbody .= "informed) may have to remove the file $icuhost{'name'}.lock manually from the\n";
	$mailbody .= "ICU server if the problem persists.\n";
    }

    if ($mailtype eq "icu_die") {
        $subject  .= "Warning: ICU had to be killed during check. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date had to be killed.\n";
	$mailbody .= "The ICU administrator has also received this email and should probably\n";
	$mailbody .= "investigate this as soon as possible.\n";
	$mailbody .= "Possible existing lock file will remain until removed manually.\n";
    }

    if ($mailtype eq "empty_report") {
        $subject  .= "AIDE found no change in the filesystem. ";
        $mailbody .= "The AIDE filesystem check of $hostname started $start_date\n";
        $mailbody .= "found no change in the filesystem.\n";
    }

    if ($mailtype eq "scp_output") {
	$subject  .= " Warning: scp gave output. " unless (defined($subject) && $subject =~ /Warning: scp gave output/);
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The scp command gave output while copying files from $icuhost{'name'}.\n";
        $mailbody .= "It may not be anything critical but you should investigate the cause.\n";
	$mailbody .= "Here is the output scp gave:\n$scp_output\n";
    }

    if ($mailtype eq "welcome") {
	$subject = $conf{'welcome_mail_subject'};
	open(WELCOME,$conf{'welcome_mail_body'}) or die("WELCOME: Could not open $conf{'welcome_mail_body'}: $!\n");
	while(<WELCOME>) {
            $_ =~ s/%hostname%/$icuhost{'name'}/g;
	    $_ =~ s/%admin_email%/$conf{'admin_email'}/g;
	    $_ =~ s/%host_basedir%/$conf{'host_basedir'}/g;
	    $mailbody .= $_;
	}
	close(WELCOME);
    }

    $debug && print STDERR "* sub get_mail ended\n";
}



sub arrcmp    # Compare two arrays and return differences. Syntax: ($added,$removed) = arrcmp(\@array1,\@array2);
{
    my ($oldref, $newref, @oldarr, @newarr, $added, $removed, $i, $j);
    $debug && print STDERR "* sub arrcmp started\n"; 

    $oldref = shift;
    $newref = shift;
    @oldarr = @$oldref;
    @newarr = @$newref;

    OUT:for ($i = 0; $i <= $#oldarr; $i++) {
	IN:for ($j = 0; $j <= $#newarr; $j++) {
            next IN unless (defined($newarr[$j]));
            if ($oldarr[$i] eq $newarr[$j]) {
                $oldarr[$i] = undef;
                $newarr[$j] = undef;
                next OUT;
            }
        }
    }

    foreach $_ (@oldarr) {	     # Remaining lines in @oldarr have been removed
	if (defined($_)) {
	    chomp($_);
            $removed .= "$_\n";
	}
    }

    foreach $_ (@newarr) {	     # Remaining lines in @newarr have been added
	if (defined($_)) {
	    chomp($_);
            $added .= "$_\n";
	}
    }

    $debug && print STDERR "* sub arrcmp returning";
    return($added,$removed);
}                            



sub tmp_dir   # Create or remove temporary directory for a host. Syntax: tmp_dir("new") or tmp_dir("remove")
{
    my ($basetmp, $error);
    $debug && print STDERR "* sub tmp_dir started\n";
    my $action = shift;

    if ($action eq "new") {  	 # Create temporary directory
	$debug && print STDERR "Creating new temporary directory ($conf{'tmp_dir'}).\n";
	$basetmp = $conf{'tmp_dir'};
	$basetmp =~ s/$icuhost{'name'}//;
	system("rm","-rf","$conf{'tmp_dir'}") if (-e "$conf{'tmp_dir'}");
	mkdir("$basetmp",0700) unless (-e "$basetmp");
	mkdir("$conf{'tmp_dir'}",0700);
    }

    if ($action eq "remove") {  # Remove temporary directory
	$debug && print STDERR "Removing temporary directory ($conf{'tmp_dir'}).\n";
	if (-e "$conf{'tmp_dir'}") {
	    ($error = `rm -rf $conf{'tmp_dir'} 2>&1`) && icu_die("$error"); 
	}
    }
    $debug && print STDERR "* sub tmp_dir ended\n";
}



sub clean_exit    # Send/print mail, write to logs and exit
{
    $debug && print STDERR "* sub clean_exit started\n";
    my ($tmpstr, $date);

    $debug && print STDERR "* sub clean_exit started\n";
    if (defined($mailbody)) {				# There is a mail to send.
	if (defined($opt_N) || defined($opt_i)) {	# Print mail to stdout if -n or -i
 	    $date = (scalar localtime);

	    notify(2,1,1,"ICU started with the -N switch - printing mail to standard output instead of sending it.")
	      unless (defined($opt_i));

	    notify(2,0,1,"Message subject: $subject");
            notify(2,0,1,"Message body:\n$mailbody");
            notify(2,0,1,"End of message body."); 
	} 
	else {
 	    send_mail($icuhost{'mailto'},$mail_admin,$subject,$mailbody);    # Send the mail
	}
    }

# When -i, Send the welcome mail and ONLY the welcome mail (other possible output was printed to stdout above) 
    if ($init_done && $conf{'send_welcome_mail'} && !defined($opt_N)) { 
	$subject  = undef;
	$mailbody = undef;
	get_mail("welcome");
	send_mail($icuhost{'mailto'},$mail_admin,$subject,$mailbody);
    }

    notify(1,1,1,"Check ended.") if (defined($opt_c));
    notify(1,1,1,"Database initialization ended.") if (defined($opt_i));
    tmp_dir("remove");
    lockfile("remove");	
    $debug && print STDERR "* sub clean_exit ended\n";
    select(STDOUT);
    exit;
}



sub notify 	# Write string to logs and also automatically to stderr if running in verbose mode. 
                # Syntax: notify(x,y,z,"string"); 
                # x=printopt, y=log_icu, z=log_host
{
    $debug && print STDERR "* sub notify started\n"; 

    my $printopt    = shift;  # print to stdout? 0=never, 1=only in verbose mode, 2=always
    my $log_icu     = shift;  # should we log to ICU.log/syslog?
    my $log_host    = shift;  # should we log to the host log?
    my ($date, $tmpstr);

    $date = (scalar localtime);
    chomp(@_);

# Print to stdout if both $verbose and $printopt are set, or if $printopt is set to 2.
    print "@_\n" if (($verbose && $printopt) || $printopt == 2);

# Write to ICU.log/syslog if $icu_log is set
    if ($log_icu) {
	$tmpstr = "$conf{'log_dir'}/ICU.log";
	open(ICULOG,">>$tmpstr") or die("ICULOG: Could not open $tmpstr for writing: $!\n");
	flock(ICULOG, LOCK_EX) or print "Warning, unable to lock ICU.log: $!\n";
	print ICULOG "$date ICU[$$] $icuhost{'name'}: @_\n";
	flock(ICULOG, LOCK_UN) or print "Warning, unable to unlock ICU.log: $!\n"; 
	close(ICULOG);

    # Same thing goes to syslog if $use_syslog is set to 1.
        system("$conf{'logger'} $conf{'logger_arguments'} -t \"ICU[$$] $icuhost{'name'}\" \"@_\n\"") 
          if ($conf{'use_syslog'});
  }


# Write to the host log if $log_host is set
    if ($log_host) {
	$tmpstr = "$conf{'log_dir'}/$icuhost{'name'}.log";
	open(HOSTLOG,">>$tmpstr") or die("HOSTLOG: Could not open $tmpstr for writing: $!\n");
	print HOSTLOG "$date ICU[$$] $icuhost{'name'}: @_\n";
	close(HOSTLOG);  
    }

    $debug && print STDERR "* sub notify ended\n";
}



sub beepbeepbeepbeep    	# Take care of the alarm. (something timed out)
{
    $debug && print STDERR "* sub beepbeepbeepbeep started\n";
    $verbose && print STDERR "Timeout.\n";
    get_mail("scp_output") if (defined($scp_output));
    notify(1,1,1,$timeout_reason);   

    if ($trying_to_connect) {	# Timeout during port check
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_port_closed'});
    }
    else {			# Timeout waiting for AIDE to finish
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_timeout'});
    }
    clean_exit;             
}



sub icu_die     # Write to logs, send mail and exit. Don't remove the lock file because a weird
		# error that needs to be investigated has probably occured when this sub is called
{
    $debug && print STDERR "* sub icu_die started\n";

    notify(1,1,1,"Fatal error: @_, Exiting...");
    get_mail("icu_die");
    $mailbody .= "Additional information:\n@_\n";

    if (defined($opt_N)) {
	print $mailbody;
    }
    else {
	send_mail($icuhost{'mailto'},1,$subject,$mailbody);
    }         

    exit;
}



sub sanity_check     # Check for required files and their permissions etc
{
    my ($output, $tmpstr, @tmparr, $perm, $date, $mailto, $required, $failbanner);
    $debug && print STDERR "* sub sanity_check started\n"; 
    $verbose && print "Initializing sanity check.\n";

    $failbanner = "Sanity check failed for command 'ICU.pl @args' executed $start_date\n";

  # Parameters that must be set in ICU config.  Exit if any parameter is missing
    my @conf_req = qw 
      ( 
	icu_basedir icu_server_name admin_email md5sum ssh scp ssh_keygen tar logger sendmail
	cp compress use_syslog send_welcome_mail mail_to_icu_admin_db_md5 mail_to_icu_admin_conf_md5
	mail_to_icu_admin_timeout allow_empty_reports compress_databases lockfile_also_notify_host_admin
	mail_to_icu_admin_port_closed continue_if_binary_changed warn_if_root getfiles_timeout
        check_timeout check_port_timeout host_basedir mail_from icuhosts log_dir bin_dir config_dir
        database_dir icu_client_installscript welcome_mail_body readme_installscript 
        welcome_mail_subject key_init_db_pub key_init_db_priv key_check_pub key_check_priv
        key_get_conf_pub key_get_conf_priv key_get_bin_pub key_get_bin_priv key_get_db_pub
        key_get_db_priv aide_check_cmd aide_init_db_cmd check_cmd init_db_cmd send_bin_cmd
        send_conf_cmd send_db_cmd get_bin_cmd get_conf_cmd get_db_cmd mail_to_icu_admin_scp_output
        mail_to_icu_admin_missing_files
      );	

    foreach $required (@conf_req) {
	unless (defined($conf{"$required"})) {
	    $_ = "* The parameter '$required' is not defined in $icu_config.\n";
            $output .= $_;    
	}	
    }

   if (defined($output)) {    # Something failed - don't even continue the sanity checks.
	print STDERR "$failbanner$output\nExiting...\n";
	exit;
   }

  # Is ICU started as root?
    check_root if ($conf{'warn_if_root'} || defined($opt_s)); 

  # Files that must exist.
    @tmparr = 
      (
         $conf{'icu_basedir'}, $conf{'tmp_dir'}, $conf{'icuhosts'}, $conf{'bin_dir'}, $conf{'config_dir'}, 
  	 $conf{'database_dir'}, $conf{'key_init_db_pub'}, $conf{'key_init_db_priv'}, $conf{'key_check_pub'},
	 $conf{'log_dir'}, $conf{'key_check_priv'}, $conf{'key_get_conf_pub'}, $conf{'key_get_conf_priv'},
	 $conf{'key_get_bin_pub'}, $conf{'key_get_bin_priv'}, $conf{'key_get_db_pub'}, $conf{'scp'},
	 $conf{'key_get_db_priv'}, $conf{'icu_client_installscript'}, $conf{'md5sum'}, $conf{'ssh'}, 	
	 $conf{'tar'}, $conf{'logger'}, $conf{'sendmail'}, $conf{'ssh_keygen'}, $conf{'readme_installscript'}
      );

    EXIST_LOOP:foreach $tmpstr (@tmparr) {
	unless (-e "$tmpstr") {  
	    $_ = "* $tmpstr doesn't exist.\n";
	    $output .= $_; 
	}
    }

  # Files that must be owned and writable by user.
    @tmparr = ($conf{'icu_basedir'}, $conf{'log_dir'}, $conf{'tmp_dir'}, $conf{'database_dir'});
    OWN_WRITE:foreach $tmpstr (@tmparr) { 
	next unless (-e "$tmpstr");   # already checked this above
	unless (-o $tmpstr) {   # not owned by user?
	    $_ = "* $tmpstr is not owned by you.\n";
	    $output .= $_;
	}
	unless (-w $tmpstr) {   # not writable by user?
	    $_ = "* $tmpstr is not writable. Use chmod u+w $tmpstr to fix it.\n"; 
	    $output .= $_; 
	    next OWN_WRITE;
	}
	if (-d $tmpstr && !-x $tmpstr) {
	    $_ = "* $tmpstr should be a writable directory and must be set +x. Use chmod u+x $tmpstr to fix it.\n";
	    $output .= $_; 
	}
    }

  # If compress_databases is set to 1, make sure the compress program exists
    if ($conf{'compress_databases'} && !-e "$conf{'compress'}") {
	$_ = "* You have set compress_databases=1 in $icu_config but the compress binary ($conf{'compress'}) was not found.\n";
	$output .= $_;  
    }

  # If $conf{'send_welcome_mail'} is set to 1, make sure the mail exists
    if ($conf{'send_welcome_mail'} &&  !-e "$conf{'welcome_mail_body'}") {
	$_ = "* You have set send_welcome_mail=1 in $icu_config but the mail $conf{'welcome_mail_body'} was not found.\n"; 
	$output .= $_;
    }

  # Do some tests on the host if user specified -r <hostname>
    if (defined($opt_r)) {
	if (get_host_entry("$hostname")) {   # Does it exist in the host config?
	    print "$hostname was found in $conf{'icuhosts'}:\nmailto=$icuhost{'mailto'}, OS=$icuhost{'os'}, architecture=$icuhost{'arch'}, port=$icuhost{'sshport'}.\n" 
              if ((defined($opt_s) && $verbose) || $debug);

	    unless (-e "$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}") {   # No binary for this OS/architecture?
	        $_ = "* No binary found for this OS/architecture.  (Should be found as $conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'})\n";
	        $output .= $_;
            }
	    unless (-e "$conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'}") {   # No config found for this OS/architecture?
		$_ = "* No configuration file found for this OS/architecture.  (Should be found as $conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'})\n";           
		$output .= $_;  
            }
        }
        else {				      # The host was not found in the host config
	    $_ = "* The host $hostname was not found in $conf{'icuhosts'}.\n";
	    $output .= $_;
	}
    }

    if (defined($opt_n)) {   # $conf{'database_dir'}/$hostname must not exist when running -n
	$tmpstr = "$conf{'database_dir'}/$hostname";
	if (-e "$tmpstr") {
	    $_ = "* The directory $tmpstr already exists. (Tarball already created?)\n";
	    $_ .= "  Delete this directory if you really want to create a new tarball.\n";
	    $output .= $_;
	} 
    }                      
	
    if (defined($opt_i)) {    # $conf{'database_dir'}/$hostname/archive must exist when running -i
	$tmpstr = "$conf{'database_dir'}/$hostname/archive";
	unless (-e "$tmpstr") {
	    $_ = "* $tmpstr doesn't exist. (Did you forget to run ICU.pl with the -n switch first?)\n"; 
	    $output .= $_;
	}
	$tmpstr = "$conf{'database_dir'}/$hostname/aide.db.current";
	if (-e "$tmpstr") {
	    $_ = "* $tmpstr already exists. (Database already initialized?)\n";
	    $_ .= "  Remove this file if you really want to initialize a new database.\n";
	    $output .= $_;      
	}
    }
		
    if (defined($opt_c)) {    # $conf{'database_dir'}/$hostname/aide.conf.current must exist when running -c.
        $tmpstr = "$conf{'database_dir'}/$hostname/aide.conf.current";
	unless (-e "$tmpstr") {
	    $_ = "* $tmpstr doesn't exist. (Did you forget to run ICU.pl with the -i switch first?)\n";
	    $output .= $_;
	}                        
	$tmpstr = "$conf{'database_dir'}/$hostname/aide.db.current";
	unless (-e "$tmpstr") {    # And aide.db must also exist.
	    $_ = "* $tmpstr doesn't exist. (Did you forget to run ICU.pl with the -i switch first?)\n";
	    $output .= $_;  
	}
    }	

  # Do a write test in the temporary directory
    if (-e "$conf{'tmp_dir'}" && -w "$conf{'tmp_dir'}" && -x "$conf{'tmp_dir'}") { 
	open(FOO,">$conf{'tmp_dir'}/foo.tmp") or $output .= "* Write test in $conf{'tmp_dir'} failed: $!\n";
	close(FOO); 
	unlink("$conf{'tmp_dir'}/foo.tmp") if (-e "$conf{'tmp_dir'}/foo.tmp");
    }

  # Do a MD5 fingerprint test - make sure we actually get a 32 char output
    if (-e "$conf{'md5sum'}" && -e "$conf{'cp'}") {
        $tmpstr = `$conf{'md5sum'} $conf{'cp'}`;
        unless ($tmpstr =~ /([\da-f]{32})/) {
 	    $_ = "* The md5sum command ($conf{'md5sum'}) failed - could not find MD5 fingerprint in output. ";
	    $_ .= "  The output was: $tmpstr";
	    $output .= $_;
	}
    }

    if (defined($output)) {    # Something failed - print errors to STDERR and logs
	$date = (scalar localtime);
	$tmpstr = "$conf{'log_dir'}/ICU.log";
	$hostname = "[no host]" unless (defined($hostname));

	if (-e "$conf{'log_dir'}" && -o "$conf{'log_dir'}") {
	    open(ICULOG,">>$tmpstr") or print STDERR "* ICULOG: Could not open $tmpstr for writing: $!\n";
	    flock(ICULOG, LOCK_EX) or print STDERR "* Warning, unable to lock ICU.log: $!\n"; 
	    print ICULOG "$date ICU[$$] $hostname: $failbanner$output"; 
	    flock(ICULOG, LOCK_UN) or print STDERR  "* Warning, unable to unlock ICU.log: $!\n";
	    close(ICULOG);

	    if (defined($icuhost{'name'})) {
	    	$tmpstr = "$conf{'log_dir'}/$icuhost{'name'}.log";
		open(HOSTLOG,">>$tmpstr") or print STDERR "HOSTLOG: Could not open $tmpstr for writing: $!\n";
    		print HOSTLOG "$date ICU[$$] $icuhost{'name'}: $failbanner$output";
    		close(HOSTLOG); 
	    }
	}

        print STDERR $failbanner;
        print STDERR "Maybe you need to edit $icu_config? Things that failed:\n$output\nExiting...\n";
 	system("$conf{'logger'} $conf{'logger_arguments'} -t \"ICU[$$] $hostname\" \"$failbanner: $output\"")
	  if (-e "$conf{'logger'}" && $conf{'use_syslog'});
	exit;
    } 

    print "Sanity check passed.\n" if (defined($opt_s) || $verbose);      
    exit if (defined($opt_s));
    $debug && print STDERR "* sub sanity_check ended\n";  
}



sub generate_keys	# Call the sub routine &generate_key with different arguments
			# for each key we need to create
{
    $debug && print STDERR "* sub generate_keys started\n";
    my (@tmparr, $tmpstr, $key_command);

    check_root if ($conf{'warn_if_root'});

  # Files/directories that must exist.
    @tmparr = ( $conf{'icu_basedir'}, $conf{'ssh_keygen'}, $conf{'tmp_dir'});
    foreach $tmpstr (@tmparr) {
         die ("$tmpstr doesn't exist.\n") unless (-e "$tmpstr");
    }                          

    system("rm $conf{'tmp_dir'}/tempkey* 2>/dev/null"); 

  # Generate init_db key
    $key_command = "command=\"$conf{'aide_init_db_cmd'}\""; 
    generate_key($conf{'key_init_db_priv'},$conf{'key_init_db_pub'},'icu_init_db',$key_command); 

  # Generate check key
    $key_command = "command=\"$conf{'aide_check_cmd'}\"";
    generate_key($conf{'key_check_priv'},$conf{'key_check_pub'},'icu_check',$key_command); 

  # Generate send_bin key
    $key_command = "command=\"$conf{'send_bin_cmd'}\"";
    generate_key($conf{'key_get_bin_priv'},$conf{'key_get_bin_pub'},'icu_get_bin',$key_command); 

  # Gernerate send_conf key
    $key_command = "command=\"$conf{'send_conf_cmd'}\"";
    generate_key($conf{'key_get_conf_priv'},$conf{'key_get_conf_pub'},'icu_get_conf',$key_command); 

  # Gernerate send_db key
    $key_command = "command=\"$conf{'send_db_cmd'}\"";
    generate_key($conf{'key_get_db_priv'},$conf{'key_get_db_pub'},'icu_get_db',$key_command); 

    system("rm $conf{'tmp_dir'}/tempkey* 2>/dev/null");

    $debug && print STDERR "* sub generate_keys ended\n"; 
    exit;
}



sub generate_key    # Syntax: generate_key(privkey,pubkey,comment,command)
{
    $debug && print STDERR "* sub generate_key started\n"; 

    my $privkey_filename = shift;
    my $pubkey_filename  = shift;
    my $comment          = shift;
    my $key_command      = shift;

  # Don't overwrite existing keys                             
    $_ = "already exists and I refuse to overwrite your existing keys.\n";
    $_ .= "Make sure your old keys are not used by any client and remove them ";
    $_ .= "if you really want to create new ones.\n";
    die("$privkey_filename $_") if (-e $privkey_filename);
    die("$pubkey_filename $_") if (-e $pubkey_filename);  

    print "Generating $privkey_filename and $pubkey_filename...\n";   
    unlink("$conf{'tmp_dir'}/tempkey")     if (-e "$conf{'tmp_dir'}/tempkey"); 
    unlink("$conf{'tmp_dir'}/tempkey.pub") if (-e "$conf{'tmp_dir'}/tempkey.pub");
    system("$conf{'ssh_keygen'} -b 1024 -C $comment -f $conf{'tmp_dir'}/tempkey -N '' 1>/dev/null 2>&1");  

    open(KEY, "$conf{'tmp_dir'}/tempkey.pub") or die "KEY: Culd not open $conf{'tmp_dir'}/tempkey.pub: $!\n";
    $_ = <KEY>;
    close(KEY);

  # Add forced command to the public key
    $key_command =~ s/%host_basedir%/$conf{'host_basedir'}/g;
    $_ = "$key_command $_";
    $pubkey_filename =~ /(.*\/)/; 
    system("mkdir -p $1") unless (-e $1);

    open(KEY, ">$pubkey_filename") or die "KEY: Could not open $pubkey_filename for writing: $!\n";
    print KEY $_;
    close(KEY);                             

    $privkey_filename =~ /(.*\/)/;
    system("mkdir -p $1") unless (-e $1); # Create key directory(/directories) if it doesn't exist.
    system("mv $conf{'tmp_dir'}/tempkey $privkey_filename") && die("\nExiting...\n");

    $debug && print STDERR "* sub generate_key ended\n"; 
}

# EOF
